from dataclasses import dataclass

from immudb.grpc import schema_pb2
from immudb.grpc import schema_pb2_grpc
from immudb.rootService import RootService
from immudb.exceptions import VerificationException

def call(service: schema_pb2_grpc.ImmuServiceStub, rs: RootService, 
            zset:bytes, seekKey:bytes, seekScore:float,
            seekAtTx:int, inclusive: bool, limit:int, desc:bool, minscore:float,
            maxscore:float, sinceTx, nowait):
    request=schema_pb2.ZScanRequest(
        set=zset,
        seekKey = seekKey,
        seekScore = seekScore,
        seekAtTx = seekAtTx,
        inclusiveSeek = inclusive,
        limit = limit,
        desc = desc,
        minScore = schema_pb2.Score(score=minscore),
        maxScore = schema_pb2.Score(score=maxscore),
        sinceTx = sinceTx,
        noWait = nowait,
        )
    msg = service.ZScan(request)
    return msg



#!/usr/bin/env python3
import hashlib
from immudb.constants import *
from immudb.exceptions import *
import immudb.store

class InclusionProof:
    def __init__(self):
        self.leaf=None
        self.width=None
        self.terms=b''

class DualProof:
    def __init__(self):
        self.sourceTxMetadata=None
        self.targetTxMetadata=None
        self.inclusionProof=None
        self.consistencyProof=None
        self.targetBlTxAlh=None
        self.lastInclusionProof=None
        self.linearProof=None
        
        
class HTree:
    def __init__(self, maxWidth:int):
        if maxWidth<1:
            return
       
        self.maxWidth=maxWidth
        lw=1
        while lw<maxWidth:
            lw=lw<<1
        height=(maxWidth-1).bit_length()+1
        self.levels=[None]*height
        for l in range(0,height):
            self.levels[l]=[None]*(lw>>l)

    def BuildWith(self, digests:list):
        if len(digests)>self.maxWidth:
            raise ErrMaxWidthExceeded
        if len(digests)==0:
            raise ErrIllegalArguments
        for i in range(0,len(digests)):
            leaf=LEAF_PREFIX+digests[i]
            self.levels[0][i]=hashlib.sha256(leaf).digest()
        l=0
        w=len(digests)
        while w>1:
            
            wn=0
            i=0
            while i+1<w:
                b=NODE_PREFIX+self.levels[l][i]+self.levels[l][i+1]
                self.levels[l+1][wn]=hashlib.sha256(b).digest()
                wn=wn+1
                i=i+2
            if w%2==1:
                self.levels[l+1][wn]=self.levels[l][w-1]
                wn=wn+1
            l+=1
            w=wn
        self.width=len(digests)
        self.root=self.levels[l][0]
    def InclusionProof(self, i):
        if i>=self.width:
            raise ErrIllegalArguments
        m=0
        n=self.width
        offset=0
        proof=InclusionProof()
        proof.leaf=i
        proof.width=self.width
        if self.width==1:
            return proof
        while True:
            d=(n-1).bit_length()
            k=1<<(d-1)
            if m<k:
                l,r=offset+k,offset+n-1
                n=k
            else:
                l,r=offset,offset+k-1
                m=m-k
                n=n-k
                offset=offset+k
            layer=(r-l).bit_length()
            index=int(l/(1<<layer))
            proof.terms=self.levels[layer][index]+proof.terms
            if n<1 or (n==1 and m==0):
                return proof
            
def InclusionProofFrom(iproof):
    h=InclusionProof()
    h.leaf=int(iproof.leaf)
    h.width=int(iproof.width)
    h.terms=immudb.store.DigestFrom(iproof.terms)
    return h

def DualProofFrom(dproof):
    dp=DualProof()
    dp.sourceTxMetadata=immudb.store.TxMetadataFrom(dproof.sourceTxMetadata)
    dp.targetTxMetadata=immudb.store.TxMetadataFrom(dproof.targetTxMetadata)
    dp.inclusionProof=immudb.store.DigestFrom(dproof.inclusionProof)
    dp.consistencyProof=immudb.store.DigestFrom(dproof.consistencyProof)
    dp.targetBlTxAlh=immudb.store.DigestFrom(dproof.targetBlTxAlh)
    dp.lastInclusionProof=immudb.store.DigestsFrom(dproof.lastInclusionProof)
    dp.linearProof=immudb.store.LinearProofFrom(dproof.linearProof) 
    return dp



from dataclasses import dataclass

from immudb.grpc import schema_pb2
from immudb.grpc import schema_pb2_grpc
from immudb.rootService import RootService
from google.protobuf.empty_pb2 import Empty

@dataclass
class dbListResponse:
    dblist: schema_pb2.DatabaseListResponse

def call(service: schema_pb2_grpc.ImmuServiceStub, rs: RootService, request: None): 
    NoRequest=Empty()
    msg = service.DatabaseList(NoRequest)
    return dbListResponse(
       dblist = msg
    )



from dataclasses import dataclass

from immudb.grpc import schema_pb2
from immudb.grpc import schema_pb2_grpc
from immudb.rootService import RootService
from immudb import datatypes

def call(service: schema_pb2_grpc.ImmuServiceStub, rs: RootService, key: bytes, offset: int, limit: int, desc: bool):
    state = rs.get()
    request = schema_pb2_grpc.schema__pb2.HistoryRequest(
                key=key,
                offset=offset,
                limit=limit,
                desc=desc,
                sinceTx=state.txId
                )
    histo = service.History(request)
    histolist=[]
    for i in histo.entries:
        histolist.append( datatypes.historyResponseItem(
            key=i.key, 
            value=i.value,
            tx=i.tx
            ))
    return histolist



import grpc
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2

from immudb import header_manipulator_client_interceptor
from immudb.handler import (batchGet, batchSet, changePassword, createUser,
                          currentRoot, databaseCreate, databaseList, databaseUse, 
                          get, listUsers, verifiedGet, verifiedSet, setValue, history, 
                          scan, reference, verifiedreference, zadd, verifiedzadd, 
                          zscan, healthcheck, txbyid, verifiedtxbyid)
from immudb.rootService import RootService
from immudb.grpc import schema_pb2_grpc
import warnings

class ImmudbClient:
    def __init__(self, immudUrl=None):
        if immudUrl is None:
            immudUrl = "localhost:3322"
        self.channel = grpc.insecure_channel(immudUrl)
        self.__stub = schema_pb2_grpc.ImmuServiceStub(self.channel)
        self.__rs = None

    def login(self, username, password, database=b"defaultdb"):
        req = schema_pb2_grpc.schema__pb2.LoginRequest(user=bytes(
            username, encoding='utf-8'), password=bytes(
                password, encoding='utf-8'
                ))
        try:
            self.__login_response = schema_pb2_grpc.schema__pb2.LoginResponse = \
                self.__stub.Login(
                    req
                )
        except ValueError as e:
            raise Exception("Attempted to login on termninated client, channel has been shutdown") from e
            
        self.__stub = self.set_token_header_interceptor(self.__login_response)
        # Select database, modifying stub function accordingly
        request = schema_pb2_grpc.schema__pb2.Database(databasename=database)
        resp = self.__stub.UseDatabase(request)
        self.__stub = self.set_token_header_interceptor(resp)

        self.init()
        return self.__login_response

    def init(self):
        rs = RootService(self.__stub)
        rs.init()
        self.__rs = rs

    def shutdown(self):
        self.channel.close()
        self.channel = None
        self.intercept_channel.close
        self.intercept_channel = None
        self.__rs = None

    def set_token_header_interceptor(self, response):
        try:
            token = response.token
        except AttributeError:
            token = response.reply.token
        self.header_interceptor = \
            header_manipulator_client_interceptor.header_adder_interceptor(
                'authorization', "Bearer "+token
            )
        try:
            self.intercept_channel = grpc.intercept_channel(
                self.channel, self.header_interceptor)
        except ValueError as e:
            raise Exception("Attempted to login on termninated client, channel has been shutdown") from e
        return schema_pb2_grpc.ImmuServiceStub(self.intercept_channel)

    @property
    def stub(self):
        return self.__stub
    
    def healthCheck(self):
        return healthcheck.call(self.__stub, self.__rs)
        

    def get(self, key: bytes):
        return get.call(self.__stub, self.__rs, key)
    
    def getValue(self, key: bytes):
        ret=get.call(self.__stub, self.__rs, key)
        if ret==None:
            return None
        return ret.value

    def set(self, key: bytes, value: bytes):
        return setValue.call(self.__stub, self.__rs, key, value)

    def safeGet(self, key: bytes):
        warnings.warn("Call to deprecated safeGet. Use verifiedGet instead",
            category=DeprecationWarning,
            stacklevel=2
            )
        return verifiedGet.call(self.__stub, self.__rs, key)
    
    def verifiedGet(self, key: bytes):
        return verifiedGet.call(self.__stub, self.__rs, key)
    
    def verifiedGetAt(self, key: bytes, atTx:int):
        return verifiedGet.call(self.__stub, self.__rs, key, atTx)

    def safeSet(self, key: bytes, value: bytes):
        warnings.warn("Call to deprecated safeSet. Use verifiedSet instead",
            category=DeprecationWarning,
            stacklevel=2
            )
        return verifiedSet.call(self.__stub, self.__rs, key, value)
    
    def verifiedSet(self, key: bytes, value: bytes):
        return verifiedSet.call(self.__stub, self.__rs, key, value)

    def getAllValues(self, keys: list):
        resp = batchGet.call(self.__stub, self.__rs, keys)
        return resp

    def getAll(self, keys: list):
        resp = batchGet.call(self.__stub, self.__rs, keys)
        return {key:value.value for key, value in resp.items()}

    def setAll(self, kv: dict):
        return batchSet.call(self.__stub, self.__rs, kv)

    def changePassword(self, user, newPassword, oldPassword):
        request = schema_pb2_grpc.schema__pb2.ChangePasswordRequest(
                    user=bytes(user, encoding='utf-8'),                     
                    newPassword=bytes(newPassword, encoding='utf-8'),
                    oldPassword=bytes(oldPassword, encoding='utf-8')
                    )
        return changePassword.call(self.__stub, self.__rs, request)
    
    def createUser(self, user, password, permission, database):
        request = schema_pb2_grpc.schema__pb2.CreateUserRequest(
                    user=bytes(user, encoding='utf-8'),
                    password=bytes(password, encoding='utf-8'),
                    permission=permission,
                    database=database
                    )
        return createUser.call(self.__stub, self.__rs, request)
    
    def listUsers(self):
        return listUsers.call(self.__stub, None)
    
    def databaseList(self):
        dbs=databaseList.call(self.__stub, self.__rs, None)
        return [x.databasename for x in dbs.dblist.databases]

    def databaseUse(self, dbName: bytes):
        request = schema_pb2_grpc.schema__pb2.Database(databasename=dbName)
        resp = databaseUse.call(self.__stub, self.__rs, request)
        # modify header token accordingly
        self.__stub = self.set_token_header_interceptor(resp)
        self.__rs = RootService(self.__stub)
        self.__rs.init()
        return resp

    def databaseCreate(self, dbName: bytes):
        request = schema_pb2_grpc.schema__pb2.Database(databasename=dbName)
        return databaseCreate.call(self.__stub, self.__rs, request)

    def currentState(self):
        return currentRoot.call(self.__stub, self.__rs, None)

    def history(self, key: bytes, offset: int, limit: int, sortorder: bool):
        return history.call(self.__stub, self.__rs, key, offset, limit, sortorder)

    def logout(self):
        self.__stub.Logout(google_dot_protobuf_dot_empty__pb2.Empty())
        self.__login_response = None
        
    def scan(self, key:bytes, prefix:bytes, desc:bool, limit:int,sinceTx:int=None):
        return scan.call(self.__stub, self.__rs, key, prefix, desc, limit, sinceTx)
    
    def setReference(self, referredkey: bytes, newkey:  bytes):
        return reference.call(self.__stub, self.__rs, referredkey, newkey)
    
    def verifiedSetReference(self, referredkey: bytes, newkey:  bytes):
        return verifiedreference.call(self.__stub, self.__rs, referredkey, newkey)
    
    def zAdd(self, zset:bytes, score:float, key:bytes, atTx:int=0):
        return zadd.call(self.__stub, self.__rs, zset, score, key, atTx)
    
    def verifiedZAdd(self, zset:bytes, score:float, key:bytes, atTx:int=0):
        return verifiedzadd.call(self.__stub, self.__rs, zset, score, key, atTx)
    
    def zScan(self, zset:bytes, seekKey:bytes, seekScore:float,
                          seekAtTx:int, inclusive: bool, limit:int, desc:bool, minscore:float,
                          maxscore:float, sinceTx=None, nowait=False):
        return zscan.call(self.__stub, self.__rs, zset, seekKey, seekScore,
                          seekAtTx, inclusive, limit, desc, minscore,
                          maxscore, sinceTx, nowait)
                          
    def txById(self, tx:int):
        return txbyid.call(self.__stub, self.__rs, tx)
    
    def verifiedTxById(self, tx:int):
        return verifiedtxbyid.call(self.__stub, self.__rs, tx)
    
        



from dataclasses import dataclass

from immudb.grpc import schema_pb2
from immudb.grpc import schema_pb2_grpc
from immudb.rootService import RootService

def call(service: schema_pb2_grpc.ImmuServiceStub, rs: RootService, key:bytes, prefix:bytes, desc:bool, limit:int, sinceTx:int):
    if sinceTx==None:
        state = rs.get()
        sinceTx=state.txId
    request = schema_pb2_grpc.schema__pb2.ScanRequest(
        seekKey=key,
        prefix=prefix,
        desc=desc,
        limit=limit,
        sinceTx=sinceTx,
        noWait=False
        )
    msg = service.Scan(request)
    ret={}
    for i in msg.entries:
        ret[i.key]=i.value
    return ret



from dataclasses import dataclass

from immudb.grpc import schema_pb2, schema_pb2_grpc
from immudb.rootService import RootService
import immudb.store
from immudb import datatypes
from immudb.exceptions import VerificationException

def call(service: schema_pb2_grpc.ImmuServiceStub, rs: RootService, refkey: bytes, key:  bytes, atTx=0):
    state = rs.get()
    req = schema_pb2_grpc.schema__pb2.ReferenceRequest(
        referencedKey = refkey,
        key=key,
        atTx=atTx,
        boundRef=atTx>0
        )
    vreq=schema_pb2_grpc.schema__pb2.VerifiableReferenceRequest(
        referenceRequest=req,
        proveSinceTx=state.txId
        )
    vtx = service.VerifiableSetReference(vreq)
    if vtx.tx.metadata.nentries !=1:
        raise VerificationException
    tx = immudb.store.TxFrom(vtx.tx)
    inclusionProof=tx.Proof(immudb.store.EncodeKey(key))
    ekv=immudb.store.EncodeReference(key, refkey, atTx)
    verifies = immudb.store.VerifyInclusion(inclusionProof, ekv.Digest(), tx.eh())
    if not verifies:
        raise VerificationException
    if state.txId == 0:
        sourceID = tx.ID
        sourceAlh = tx.Alh
    else:
        sourceID = state.txId
        sourceAlh = immudb.store.DigestFrom(state.txHash)
    targetID = tx.ID
    targetAlh = tx.Alh
    verifies = immudb.store.VerifyDualProof(
            immudb.htree.DualProofFrom(vtx.dualProof),
            sourceID,
            targetID,
            sourceAlh,
            targetAlh,
    )
    if not verifies:
        raise VerificationException
    state=schema_pb2.ImmutableState(
            txId=      targetID,
            txHash=    targetAlh,
            signature= vtx.signature,
            )
    rs.set(state)
    return datatypes.SetResponse(
        id=vtx.tx.metadata.id,
        prevAlh=vtx.tx.metadata.prevAlh,
        timestamp=vtx.tx.metadata.ts,
        eh=vtx.tx.metadata.eH,
        blTxId=vtx.tx.metadata.blTxId,
        blRoot=vtx.tx.metadata.blRoot,
        verified=True,
    )





from dataclasses import dataclass

from immudb.grpc import schema_pb2, schema_pb2_grpc
from immudb.rootService import RootService

def call(service: schema_pb2_grpc.ImmuServiceStub, rs: RootService, refkey: bytes, key:  bytes):
    request = schema_pb2_grpc.schema__pb2.ReferenceRequest(
        referencedKey = refkey,
        key=key,
        atTx=0,
        boundRef=False
        )
    msg = service.SetReference(request)
    return msg



from immudb import constants
from immudb.grpc import schema_pb2, schema_pb2_grpc
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2


class RootService:
    
    def __init__(self, service: schema_pb2_grpc.ImmuServiceStub):
        self.__cache = None
        self.__service = service

    def init(self):
        state = self.__service.CurrentState(google_dot_protobuf_dot_empty__pb2.Empty())
        self.__cache=state

    def get(self) -> schema_pb2.ImmutableState:
        if self.__cache==None:
            self.init()
        return self.__cache

    def set(self, root: schema_pb2.ImmutableState):
        self.__cache=root



from dataclasses import dataclass

@dataclass
class SetResponse:
    id: int
    prevAlh: bytes 
    timestamp: int
    eh: bytes
    blTxId: int
    blRoot: bytes
    verified: bool

@dataclass
class SafeGetResponse:
    id: int
    key: bytes
    value: bytes
    timestamp: int
    verified: bool
    refkey: bytes

@dataclass
class CurrentRootResponse:
    id: int
    hash: bytes

@dataclass
class historyResponseItem:
    key: bytes
    value: bytes
    tx: int

@dataclass
class GetResponse:
    tx: int
    key: bytes
    value: bytes




from dataclasses import dataclass

from immudb.grpc import schema_pb2
from immudb.grpc import schema_pb2_grpc
from immudb.rootService import RootService
from immudb.exceptions import VerificationException
from immudb import datatypes
import immudb.store

def call(service: schema_pb2_grpc.ImmuServiceStub, rs: RootService, zset:bytes, score:float, key:bytes, atTx:int=0):
    state = rs.get()
    request=schema_pb2.VerifiableZAddRequest(
        zAddRequest=schema_pb2.ZAddRequest(
            set=      zset,
            score=    score,
            key=      key,
            atTx=     atTx,
        ),
        proveSinceTx=state.txId
        )
    vtx = service.VerifiableZAdd(request)
    if vtx.tx.metadata.nentries!=1:
        raise VerificationException
    tx = immudb.store.TxFrom(vtx.tx)
    ekv = immudb.store.EncodeZAdd(zset, score, key, atTx)
    inclusionProof=tx.Proof(ekv.key)
    verifies = immudb.store.VerifyInclusion(inclusionProof, ekv.Digest(), tx.eh())
    if not verifies:
        raise VerificationException
    if tx.eh() != immudb.store.DigestFrom(vtx.dualProof.targetTxMetadata.eH):
        raise VerificationException
    if state.txId == 0:
        sourceID = tx.ID
        sourceAlh = tx.Alh
    else:
        sourceID = state.txId
        sourceAlh = immudb.store.DigestFrom(state.txHash)
    targetID = tx.ID
    targetAlh = tx.Alh
    verifies = immudb.store.VerifyDualProof(
            immudb.htree.DualProofFrom(vtx.dualProof),
            sourceID,
            targetID,
            sourceAlh,
            targetAlh,
    )
    if not verifies:
        raise VerificationException

    state=schema_pb2.ImmutableState(
            txId=      targetID,
            txHash=    targetAlh,
            signature= vtx.signature,
            )
    rs.set(state)
    return datatypes.SetResponse(
        id=vtx.tx.metadata.id,
        prevAlh=vtx.tx.metadata.prevAlh,
        timestamp=vtx.tx.metadata.ts,
        eh=vtx.tx.metadata.eH,
        blTxId=vtx.tx.metadata.blTxId,
        blRoot=vtx.tx.metadata.blRoot,
        verified=True,
    )



from time import time

from immudb.grpc import schema_pb2
from immudb.grpc import schema_pb2_grpc
from immudb.rootService import RootService
from immudb import constants, datatypes

import immudb.store
#import base64

def call(service: schema_pb2_grpc.ImmuServiceStub, rs: RootService, key: bytes, value:bytes):
    state = rs.get()
    # print(base64.b64encode(state.SerializeToString()))
    kv = schema_pb2.KeyValue(key=key, value=value)
    rawRequest = schema_pb2.VerifiableSetRequest(
        setRequest = schema_pb2.SetRequest(KVs=[kv]),
        proveSinceTx= state.txId,
    )
    verifiableTx = service.VerifiableSet(rawRequest)
    # print(base64.b64encode(verifiableTx.SerializeToString()))
    tx=immudb.store.TxFrom(verifiableTx.tx)
    inclusionProof=tx.Proof(constants.SET_KEY_PREFIX+key)
    ekv=immudb.store.EncodeKV(key, value)
    verifies=immudb.store.VerifyInclusion(inclusionProof, ekv.Digest(), tx.eh())
    if not verifies:
        raise VerificationException
    if tx.eh() != immudb.store.DigestFrom(verifiableTx.dualProof.targetTxMetadata.eH):
        raise VerificationException
    if state.txId == 0:
        sourceID = tx.ID
        sourceAlh = tx.Alh
    else:
        sourceID = state.txId
        sourceAlh = immudb.store.DigestFrom(state.txHash)
    targetID = tx.ID
    targetAlh = tx.Alh

    verifies = immudb.store.VerifyDualProof(
            immudb.htree.DualProofFrom(verifiableTx.dualProof),
            sourceID,
            targetID,
            sourceAlh,
            targetAlh,
    )
    if not verifies:
        raise VerificationException
    state=schema_pb2.ImmutableState(
            txId=      targetID,
            txHash=    targetAlh,
            signature= verifiableTx.signature,
            )
    rs.set(state)
    return datatypes.SetResponse(
        id=verifiableTx.tx.metadata.id,
        prevAlh=verifiableTx.tx.metadata.prevAlh,
        timestamp=verifiableTx.tx.metadata.ts,
        eh=verifiableTx.tx.metadata.eH,
        blTxId=verifiableTx.tx.metadata.blTxId,
        blRoot=verifiableTx.tx.metadata.blRoot,
        verified=verifies,
    )
    



from dataclasses import dataclass

from immudb.grpc import schema_pb2
from immudb.grpc import schema_pb2_grpc
from immudb.rootService import RootService
from google.protobuf.empty_pb2 import Empty

@dataclass
class createUserResponse:
    reply: Empty

def call(service: schema_pb2_grpc.ImmuServiceStub, rs: RootService, request: schema_pb2.CreateUserRequest): 
    
    msg = service.CreateUser(request)
    return createUserResponse(
       reply = msg
    )



from dataclasses import dataclass

from immudb.grpc import schema_pb2
from immudb.grpc import schema_pb2_grpc
from immudb.rootService import RootService
from google.protobuf.empty_pb2 import Empty

@dataclass
class listUsersResponse:
    userlist: schema_pb2.UserList
    
def call(service: schema_pb2_grpc.ImmuServiceStub, request: None):
    NoRequest=Empty()
    msg = service.ListUsers(NoRequest)
    return listUsersResponse(
       userlist = msg
    )
    



from immudb.grpc import schema_pb2_grpc
from google.protobuf import empty_pb2
from immudb.rootService import RootService

def call(service: schema_pb2_grpc.ImmuServiceStub, rs: RootService):
    resp= service.Health(empty_pb2.Empty())
    return resp.status



from dataclasses import dataclass

from immudb.grpc import schema_pb2
from immudb.grpc import schema_pb2_grpc
from immudb.rootService import RootService
from immudb.exceptions import VerificationException
from immudb import datatypes

def call(service: schema_pb2_grpc.ImmuServiceStub, rs: RootService, zset:bytes, score:float, key:bytes, atTx:int=0):
    request=schema_pb2.ZAddRequest(
            set=      zset,
            score=    score,
            key=      key,
            atTx=     atTx,
            boundRef= atTx > 0,
        )
    msg = service.ZAdd(request)
    if msg.nentries!=1:
        raise VerificationException
    return datatypes.SetResponse(
        id=msg.id,
        prevAlh=msg.prevAlh,
        timestamp=msg.ts,
        eh=msg.eH,
        blTxId=msg.blTxId,
        blRoot=msg.blRoot,
        verified=False,
    )



from dataclasses import dataclass

from immudb.grpc import schema_pb2
from immudb.grpc import schema_pb2_grpc
from immudb.rootService import RootService
from immudb import datatypes

def call(service: schema_pb2_grpc.ImmuServiceStub, rs: RootService, tx:int):
    request=schema_pb2.TxRequest(
        tx=tx
        )
    try:
        msg = service.TxById(request)
    except Exception as e:
        if hasattr(e,'details') and e.details()=='tx not found':
            return None
        raise e
    ret=[]
    for t in msg.entries:
        ret.append(t.key[1:])
    return ret



from immudb.grpc import schema_pb2
from immudb.grpc import schema_pb2_grpc
from immudb.rootService import RootService
from immudb import datatypes

def call(service: schema_pb2_grpc.ImmuServiceStub, rs: RootService, request: None):
    root = rs.get()
    return datatypes.CurrentRootResponse(id=root.txId, hash=root.txHash)



from dataclasses import dataclass

from immudb.grpc import schema_pb2
from immudb.grpc import schema_pb2_grpc
from immudb.rootService import RootService

@dataclass
class batchElement:
    tx: int
    key: bytes
    value: bytes

def call(service: schema_pb2_grpc.ImmuServiceStub, rs: RootService, keys: list):
    request = schema_pb2.KeyListRequest(
        keys = keys#[schema_pb2_grpc.schema__pb2.Key(key=k) for k in keys]
    )
    msg = service.GetAll(request)
    ret={}
    for i in msg.entries:
        element=batchElement(
            tx=i.tx,
            key=i.key,
            value=i.value
            )
        ret[i.key]=element
    return ret



#import hashlib
#import struct
#
#from immudb import constants
#
#
#def digest(index: int, key: bytes, value: bytes) -> bytes:
#    c = bytearray()
#    c.append(constants.LEAF_PREFIX)
#    c.extend(struct.pack('>Q', index))
#    c.extend(struct.pack('>Q', len(key)))
#    c.extend(key)
#    c.extend(value)
#    return hashlib.sha256(c).digest()



from dataclasses import dataclass

from immudb.grpc import schema_pb2
from immudb.grpc import schema_pb2_grpc
from immudb.rootService import RootService
from immudb import datatypes, htree, store, exceptions

def call(service: schema_pb2_grpc.ImmuServiceStub, rs: RootService, tx:int):
    state = rs.get()
    request=schema_pb2.VerifiableTxRequest(
        tx=tx,
        proveSinceTx=state.txId
        )
    try:
        vtx = service.VerifiableTxById(request)
    except Exception as e:
        if hasattr(e,'details') and e.details()=='tx not found':
            return None
        raise e
    dualProof = htree.DualProofFrom(vtx.dualProof)
    if state.txId <= vtx.tx.metadata.id:
        sourceid=state.txId
        sourcealh=store.DigestFrom(state.txHash)
        targetid=vtx.tx.metadata.id
        targetalh=dualProof.targetTxMetadata.alh()
    else:
        sourceid=vtx.tx.metadata.id
        sourcealh=dualProof.sourceTxMetadata.alh()
        targetid=state.txId
        targetalh=store.DigestFrom(state.txHash)
    verifies=store.VerifyDualProof(
        dualProof,
        sourceid,
        targetid,
        sourcealh,
        targetalh)
    if not verifies:
        raise exceptions.VerificationException
    state=schema_pb2.ImmutableState(
            txId=      targetid,
            txHash=    targetalh,
            signature= vtx.signature,
            )
    rs.set(state)

    ret=[]
    for t in vtx.tx.entries:
        ret.append(t.key[1:])
    return ret



from immudb import datatypes
from immudb.grpc import schema_pb2, schema_pb2_grpc
from immudb.rootService import RootService


    
def call(service: schema_pb2_grpc.ImmuServiceStub, rs: RootService, kv: dict):
    request=schema_pb2.SetRequest(
        KVs=[schema_pb2.KeyValue(key=key, value=value) for key,value in kv.items()]
        )
    msg = service.Set(request)
    return datatypes.SetResponse(
        id=msg.id,
        prevAlh=msg.prevAlh,
        timestamp=msg.ts,
        eh=msg.eH,
        blTxId=msg.blTxId,
        blRoot=msg.blRoot,
        verified=False,
    )



from immudb.grpc import schema_pb2
from immudb.grpc import schema_pb2_grpc
from immudb.rootService import RootService
from immudb import constants, item, htree, store, datatypes
from immudb.exceptions import VerificationException


import sys
def call(service: schema_pb2_grpc.ImmuServiceStub, rs: RootService, requestkey: bytes, atTx:int=None):
    state = rs.get()
    if atTx==None:
        req = schema_pb2.VerifiableGetRequest(
            keyRequest= schema_pb2.KeyRequest(key=requestkey),
            proveSinceTx= state.txId
            )
    else:
        req = schema_pb2.VerifiableGetRequest(
            keyRequest= schema_pb2.KeyRequest(key=requestkey, atTx=atTx),
            proveSinceTx= state.txId
            )
    ventry=service.VerifiableGet(req)
    inclusionProof = htree.InclusionProofFrom(ventry.inclusionProof)
    dualProof = htree.DualProofFrom(ventry.verifiableTx.dualProof)
    
    if ventry.entry.referencedBy==None or ventry.entry.referencedBy.key==b'':
        vTx=ventry.entry.tx
        kv=store.EncodeKV(requestkey, ventry.entry.value)
    else:
        vTx = ventry.entry.referencedBy.tx
        kv=store.EncodeReference(ventry.entry.referencedBy.key, ventry.entry.key, ventry.entry.referencedBy.atTx) 
        
    if state.txId <= vTx:
        eh=store.DigestFrom(ventry.verifiableTx.dualProof.targetTxMetadata.eH)
        sourceid=state.txId
        sourcealh=store.DigestFrom(state.txHash)
        targetid=vTx
        targetalh=dualProof.targetTxMetadata.alh()
    else:
        eh=store.DigestFrom(ventry.verifiableTx.dualProof.sourceTxMetadata.eH)
        sourceid=vTx
        sourcealh=dualProof.sourceTxMetadata.alh()
        targetid=state.txId
        targetalh=store.DigestFrom(state.txHash)
        
    verifies = store.VerifyInclusion(inclusionProof,kv.Digest(),eh)
    if not verifies:
        raise VerificationException
    verifies=store.VerifyDualProof(
        dualProof,
        sourceid,
        targetid,
        sourcealh,
        targetalh)
    if not verifies:
        raise VerificationException
    state=schema_pb2.ImmutableState(
            txId=      targetid,
            txHash=    targetalh,
            signature= ventry.verifiableTx.signature,
            )
    rs.set(state)
    if ventry.entry.referencedBy!=None and ventry.entry.referencedBy.key!=b'':
        refkey=ventry.entry.referencedBy.key
    else:
        refkey=None
    return datatypes.SafeGetResponse(
        id=vTx,
        key=ventry.entry.key,
        value=ventry.entry.value,
        timestamp=ventry.verifiableTx.tx.metadata.ts,
        verified=verifies,
        refkey=refkey
        )



from dataclasses import dataclass

from immudb.grpc import schema_pb2, schema_pb2_grpc
from immudb.rootService import RootService
from google.protobuf.empty_pb2 import Empty

@dataclass
class changePasswordResponse:
    reply: Empty

def call(service: schema_pb2_grpc.ImmuServiceStub, rs: RootService, request: schema_pb2.ChangePasswordRequest): 
    
    msg = service.ChangePassword(request)
    return changePasswordResponse(
       reply = msg
    )



from immudb.grpc import schema_pb2
from immudb.grpc import schema_pb2_grpc
from immudb.rootService import RootService
from immudb import datatypes

def call(service: schema_pb2_grpc.ImmuServiceStub, rs: RootService, key: bytes, value: bytes): 
    request=schema_pb2.SetRequest(
        KVs=[schema_pb2.KeyValue(key=key, value=value)]
        )
    msg = service.Set(request)
    return datatypes.SetResponse(
        id=msg.id,
        prevAlh=msg.prevAlh,
        timestamp=msg.ts,
        eh=msg.eH,
        blTxId=msg.blTxId,
        blRoot=msg.blRoot,
        verified=False,
    )



from dataclasses import dataclass

from immudb.grpc import schema_pb2
from immudb.grpc import schema_pb2_grpc
from immudb.rootService import RootService
from immudb import datatypes

def call(service: schema_pb2_grpc.ImmuServiceStub, rs: RootService, key:bytes):
    request=schema_pb2.KeyRequest(
        key=key
        )
    try:
        msg = service.Get(request)
    except Exception as e:
        if hasattr(e,'details') and e.details()=='key not found':
            return None
        raise e
    
    return datatypes.GetResponse(
        tx=msg.tx,
        key=msg.key,
        value=msg.value
        )
